/* LatVis
 * Copyright (C) 2002 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "diagramarea.h"
#include "diagramwindow.h"
#include "diagrampopupmenu.h"
#include "dialogs.h"
#include "diagram_graphic.h"
#include <algorithm>
#include <stdio.h>

#define QT_FULL_ANGLE 5760

void
DiagramArea::coords_diagram_to_area(int &x, int &y, int min_x, int min_y, 
				    int w, int h)
{
  x = x - min_x + ((width() - w) >> 1);
  y = h - (y - min_y) + ((height() - h) >> 1);
}

void
DiagramArea::coords_area_to_diagram(int &x, int &y, int min_x, int min_y, 
				    int w, int h)
{
  x = x - ((width() - w) >> 1) + min_x;
  y = h - (y - ((height() - h) >> 1) - min_y);
}

void
DiagramArea::drawVertex(QPainter *painter, int x, int y, string label, 
			bool selected, bool marked, bool hidden_label)
{
  int point_size_2;
  QColor col = color;
  QFontMetrics fm(text_font);
  QRect rect;
  
  point_size_2 = diagram_vertex_size / 2;
  // color
  if (selected) col.setRgb(col.red() + selected_color.red(),
			   col.green() + selected_color.green(),
			   col.blue() + selected_color.blue());
  if (marked) col.setRgb(col.red() + marked_color.red(),
			 col.green() + marked_color.green(),
			 col.blue() + marked_color.blue());
  // draw vertex
  painter->setPen(NoPen);
  painter->setBrush(col);
  painter->drawPie(x - point_size_2, y - point_size_2, diagram_vertex_size, 
		   diagram_vertex_size, 0, QT_FULL_ANGLE);
  // draw label
  // TODO: will be an option in preferences
  if (!hidden_label) {
    painter->setPen(col);
    painter->setFont(text_font);
    rect = fm.boundingRect(label.c_str());
    painter->setBrush(back_color);
    painter->drawRect(x + diagram_vertex_size - 2, y - rect.height() - 2,
		      rect.width() + 6, rect.height() + 4);
    painter->drawText(x + diagram_vertex_size, y, label.c_str());
  }
}

void
DiagramArea::drawDiagram()
{
  set<Vertex> verteces;
  set<Vertex>::iterator iter_s1, iter_s2;
  Element *elem;
  set<Element>::iterator iter_se;
  int vx1, vy1, vx2, vy2;
  Coord_t x1, y1, x2, y2, w, h;
  Vertex vert(NoElement);

  orderedset->getDiagram()->getBox(x1, y1, x2, y2);
  w = x2 - x1; h = y2 - y1;
  // draw background
  pixmap_painter.setBrush(back_color);
  pixmap_painter.setPen(NoPen);
  pixmap_painter.drawRect(0, 0, width(), height());
  orderedset->getDiagram()->getVerteces(verteces);
  pixmap_painter.setPen(color);
  for (iter_s1 = verteces.begin(); iter_s1 != verteces.end(); iter_s1++) {
    vx1 = (*iter_s1).x; vy1 = (*iter_s1).y;
    coords_diagram_to_area(vx1, vy1, x1, y1, w, h);
    for (iter_s2 = verteces.begin(); iter_s2 != verteces.end(); iter_s2++) {
      if (*iter_s1 == *iter_s2) continue;
      if (orderedset->getDiagram()->isEdge(*iter_s1, *iter_s2)) {
	vx2 = (*iter_s2).x; vy2 = (*iter_s2).y;
	coords_diagram_to_area(vx2, vy2, x1, y1, w, h);
	// draw edge
	pixmap_painter.drawLine(vx1, vy1, vx2, vy2);
      }
    }
  }
  for (iter_s1 = verteces.begin(); iter_s1 != verteces.end(); iter_s1++) {
    vert = *iter_s1;
    vx1 = vert.x; vy1 = vert.y;
    coords_diagram_to_area(vx1, vy1, x1, y1, w, h);
    elem = orderedset->getDiagram()->getElement(vert);
    if (elem) {
      iter_se = selected_elements.find(*elem);
      if (iter_se != selected_elements.end())
	vert.properties |= Vertex::SELECTED;
      else vert.properties &= (unsigned long)~Vertex::SELECTED;
      iter_se = marked_elements.find(*elem);
      if (iter_se != marked_elements.end())
	vert.properties |= Vertex::MARKED;
      else vert.properties &= (unsigned long)~Vertex::MARKED;
      orderedset->getDiagram()->updateVertex(vert);
      drawVertex(&pixmap_painter, vx1, vy1, elem->name,
		 vert.properties & Vertex::SELECTED,
		 vert.properties & Vertex::MARKED, 
		 vert.properties & Vertex::HIDDEN_LABEL);
    }
  }
  orderedset->setUpdated();
}

void
DiagramArea::flush(int x, int y, int w, int h)
{
  //QPixmap null_pixmap;
  
  back_pixmap_painter.setBrush(back_color);
  back_pixmap_painter.setPen(NoPen);
  back_pixmap_painter.drawRect(x, y, w, h);
  back_pixmap_painter.drawPixmap(x, y, pixmap, x, y, w, h);
  QPainter(this).drawPixmap(0, 0, back_pixmap, 0, 0, 
			    back_pixmap.width(), back_pixmap.height());
  //setBackgroundPixmap(back_pixmap);
  //erase(x, y, w, h);
  //setBackgroundPixmap(null_pixmap);
}

void
DiagramArea::resizeEvent(QResizeEvent *ev)
{
  pixmap_painter.end(); back_pixmap_painter.end();
  pixmap.resize(ev->size().width(), ev->size().height());
  back_pixmap.resize(ev->size().width(), ev->size().height());
  pixmap_painter.begin(&pixmap); back_pixmap_painter.begin(&back_pixmap);
  drawDiagram();
  QWidget::resizeEvent(ev);
}
 
void
DiagramArea::paintEvent(QPaintEvent *ev)
{
  flush(ev->rect().x(), ev->rect().y(), ev->rect().width(), 
	ev->rect().height());
  QWidget::paintEvent(ev);
}

void
DiagramArea::mousePressEvent(QMouseEvent *ev)
{
  int x, y, x0, y0;
  Vertex *vert_p;
  Vertex vert(NoElement);
  set<Element>::iterator iter_s;
  NewElementDialog *new_dialog;
  string name;
  Element *elem = NULL;
  Coord_t x1, y1, x2, y2, w, h;
  list<set<Element> >::iterator iter_l;
  
  x = ev->x(); y = ev->y(); x0 = x; y0 = y;
  orderedset->getDiagram()->getBox(x1, y1, x2, y2);
  w = x2 - x1; h = y2 - y1;
  coords_area_to_diagram(x, y, x1, y1, w, h);
  if (ev->button() == LeftButton) {
    vert_p = orderedset->getDiagram()->
      getVertexInBox(x, y, diagram_vertex_size / 2);
    if (vert_p) elem = orderedset->getDiagram()->getElement(*vert_p);
    if (popupmenu->getMode() == DiagramPopupMenu::MODE_SELECT) {
      drag_vertex = vert_p;
      drag_old_diagram_coord_x = x; drag_old_diagram_coord_y = y;
      if (elem) {
	if (ev->state() & ShiftButton) {
	  iter_s = selected_elements.find(*elem);
	  if (iter_s == selected_elements.end()) 
	    selected_elements.insert(*elem);
	  else selected_elements.erase(iter_s);
	}
	mode_elements.clear();
      }
    }
    else if (popupmenu->getMode() == DiagramPopupMenu::MODE_NEW) {
      new_dialog = new NewElementDialog;
      vert = Vertex(NoElement, NoElement, x, y);
      new_dialog->exec();
      name = new_dialog->getElementName();
      delete new_dialog;
      if (!name.empty()) {
	orderedset->save_UndoRedo();
	orderedset->newElement(name, &vert);
      }
      mode_elements.clear();
    }
    else if (popupmenu->getMode() == DiagramPopupMenu::MODE_REMOVE) {
      if (elem) {
	orderedset->save_UndoRedo();
	orderedset->removeElement(*elem);
	selected_elements.erase(*elem);
	for (iter_l = path_elems.begin(); iter_l != path_elems.end();
	     iter_l++)
	  (*iter_l).erase(*elem);
	selectpath_dialog->setPaths(path_elems);
	mode_elements.clear();
      }
    }
    else if (popupmenu->getMode() == DiagramPopupMenu::MODE_COVER) {
      if (elem) {
	mode_elements.push_back(*elem);
	if (mode_elements.size() == 2) {
	  orderedset->save_UndoRedo();
	  orderedset->orderElements(mode_elements.front(),
				    mode_elements.back());
	  mode_elements.clear();
	}
      }
    }
    else if (popupmenu->getMode() == DiagramPopupMenu::MODE_REMOVE_COVER) {
      if (elem) {
	mode_elements.push_back(*elem);
	if (mode_elements.size() == 2) {
	  orderedset->save_UndoRedo();
	  orderedset->removeOrder(mode_elements.front(), 
				  mode_elements.back());
	  mode_elements.clear();
	}
      }
    }
    else if (popupmenu->getMode() == DiagramPopupMenu::MODE_MARK_UPPER) {
      if (elem) {
	marked_elements.clear();
	orderedset->upperElements(marked_elements, *elem);
	mode_elements.clear();
      }
    }
    else if (popupmenu->getMode() == DiagramPopupMenu::MODE_MARK_LOWER) {
      if (elem) {
	marked_elements.clear();
	orderedset->lowerElements(marked_elements, *elem);
	mode_elements.clear();
      }
    }
    else if (popupmenu->getMode() == DiagramPopupMenu::MODE_MARK_PATHS) {
      if (elem) {
	mode_elements.push_back(*elem);
	if (mode_elements.size() == 2) {
	  path_elems.clear();
	  orderedset->getPaths(path_elems, mode_elements.front(), 
			       mode_elements.back());
	  mode_elements.clear();
	  marked_elements.clear();
	  selectpath_dialog->setPaths(path_elems);
	  if (path_elems.size() > 1) selectpath_dialog->exec();
	  else if (!path_elems.empty()) {
	    marked_elements = path_elems.front();
	  }
	}
      }
    }
    drawDiagram();
    flush(0, 0, width(), height());
    window->setTitle();
  }
  else if (ev->button() == RightButton) {
    mode_elements.clear();
    popupmenu->popup(x, y, x0, y0);
  }
}

void
DiagramArea::mouseMoveEvent(QMouseEvent *ev)
{
  int x, y;
  Coord_t dx, dy;
  Vertex *vert;
  set<Element>::iterator iter_s;
  Coord_t x1, y1, x2, y2, w, h;
  
  x = ev->x(); y = ev->y();
  if ((ev->state() & LeftButton) &&
      (x > diagram_border_x) && (x < (width() - diagram_border_x)) &&
      (y > diagram_border_y) && (y < (height() - diagram_border_y))) {
    orderedset->getDiagram()->getBox(x1, y1, x2, y2);
    w = x2 - x1; h = y2 - y1;
    coords_area_to_diagram(x, y, x1, y1, w, h);
    dx = x - drag_old_diagram_coord_x; dy = y - drag_old_diagram_coord_y;
    if ((!(ev->state() & ShiftButton)) && (drag_vertex != NULL)) {
      drag_vertex->x += dx; drag_vertex->y += dy;
      drag_vertex = orderedset->getDiagram()->updateVertex(*drag_vertex);
    }
    if ((ev->state() & ShiftButton) && (!(selected_elements.empty())))
      for (iter_s = selected_elements.begin(); 
	   iter_s != selected_elements.end(); iter_s++) {
	vert = orderedset->getDiagram()->getVertex(*iter_s);
	vert->x += dx; vert->y += dy;
	orderedset->getDiagram()->updateVertex(*vert);
      }
    drag_old_diagram_coord_x = x; drag_old_diagram_coord_y = y;
    drawDiagram();
    flush(0, 0, width(), height());
    window->setTitle();
  }
}

DiagramArea::DiagramArea(QWidget *parent, DiagramWindow *win, 
			 DiagramPopupMenu *pmenu, OrderedSets *oss, 
			 OrderedSet *os, int min_width, int min_height)
  : QWidget(parent),
    diagram_vertex_size(10),
    diagram_border_x(20), diagram_border_y(20),
    color("black"), back_color("white"), selected_color("green"),
    marked_color("blue"), text_font("fixed"),
    pixmap(1, 1), back_pixmap(1, 1),
    pixmap_painter(&pixmap), back_pixmap_painter(&back_pixmap)
{
  Coord_t x1, y1, x2, y2, w, h;

  orderedsets = oss;
  orderedset = os;
  window = win;
  popupmenu = pmenu;
  drag_vertex = NULL;
  min_diagram_width = min_width; min_diagram_height = min_height;
  setBackgroundMode(NoBackground);
  orderedset->getDiagram()->getBox(x1, y1, x2, y2);
  w = x2 - x1; h = y2 - y1;
  // size
  if (w < min_diagram_width) w = min_diagram_width;
  if (h < min_diagram_height) h = min_diagram_height;
  resize(w + 2 * diagram_border_x, h + 2 * diagram_border_y);
  setMinimumSize(size());
  selectpath_dialog = new SelectPathDialog(this);
  show();
}

void
DiagramArea::update()
{
  Coord_t x1, y1, x2, y2, w, h;

  orderedset->getDiagram()->getBox(x1, y1, x2, y2);
  w = x2 - x1; h = y2 - y1;
  // size
  if (w < min_diagram_width) w = min_diagram_width;
  if (h < min_diagram_height) h = min_diagram_height;
  // TODO: Qt doesn't update (show) scrollbars when the vieport contents's size change if they are not shown before size change!
  if ((w > width()) || (h > height())) {
    resize(w + 2 * diagram_border_x, h + 2 * diagram_border_y);
    setMinimumSize(size());
  }
  drawDiagram();
  flush(0, 0, width(), height());
  window->setTitle();
}

void
DiagramArea::removeSelectedElements()
{
  list<set<Element> >::iterator iter_l;
  set<Element>::iterator iter_s;

  orderedset->save_UndoRedo();
  for (iter_s = selected_elements.begin(); iter_s != selected_elements.end();
       iter_s++) {
    orderedset->removeElement(*iter_s);
    for (iter_l = path_elems.begin(); iter_l != path_elems.end();
	 iter_l++)
      (*iter_l).erase(*iter_s);
  }
  selectpath_dialog->setPaths(path_elems);
  selected_elements.clear();
}

void 
DiagramArea::invertSelectedElements()
{
  set<Element> elems, invert;
  
  orderedset->getElements(elems);
  set_difference(elems.begin(), elems.end(), selected_elements.begin(),
		 selected_elements.end(), insert_iterator<set<Element> >
                 (invert, invert.begin()));
  selected_elements = invert;
}
