/* LatVis
 * Copyright (C) 2002 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "orderdata_tree.h"
#include <stdio.h>

bool
OrderData_Tree::isElement(ElementId x)
{
  Pair p;
  
  p.first = x; p.second = NoElement;
  if (pairs_first.find(&p) != pairs_first.end()) return true;
  return false;
}

bool
OrderData_Tree::insertPair(ElementId x, ElementId y)
{
  Pair *p;
  
  p = new Pair; p->first = x; p->second = y;
  pairs_first.insert(p); pairs_second.insert(p);
  return true;
}

bool
OrderData_Tree::erasePair(ElementId x, ElementId y)
{
  Pair p;
  set<Pair *, Pair_compare_first>::iterator iter_spf;
  set<Pair *, Pair_compare_second>::iterator iter_sps;
  
  p.first = x; p.second = y;
  iter_spf = pairs_first.find(&p);
  if (iter_spf == pairs_first.end()) return false;
  iter_sps = pairs_second.find(&p);
  delete *iter_spf;
  pairs_first.erase(iter_spf); pairs_second.erase(iter_sps);
  return true;
}

bool
OrderData_Tree::lessTransitive(ElementId x, ElementId y, bool transitivity)
{
  /*
    for all (x, z) do recursively for (z, y)
  */
  Pair p;
  set<Pair *, Pair_compare_first>::iterator iter_s;
  
  p.first = x; p.second = y;
  iter_s = pairs_first.find(&p);
  if (iter_s != pairs_first.end()) return true;
  if (!transitivity) return false;
  p.second = NoElement;
  for (iter_s = pairs_first.find(&p), iter_s++;
       (iter_s != pairs_first.end()) && ((*iter_s)->first == x); iter_s++)
    if (lessTransitive((*iter_s)->second, y)) return true;
  return false;
}

void
OrderData_Tree::minimal_maximal(set<ElementId> &s, set<ElementId> ss, 
				bool minimal)
{
  /*
    minimal m: does not exist (x, m), x != m, x, m from ss
    maximal M: does not exist (M, x), M != x, M, x from ss
  */
  set<ElementId>::iterator iter_s1, iter_s2;
  
  for (iter_s1 = ss.begin(); iter_s1 != ss.end(); iter_s1++) {
    for (iter_s2 = ss.begin(); iter_s2 != ss.end(); iter_s2++) {
      if (*iter_s1 == *iter_s2) continue;
      if (minimal) { if (lessTransitive(*iter_s2, *iter_s1)) break; }
      else if (lessTransitive(*iter_s1, *iter_s2)) break;
    }
    if (iter_s2 == ss.end()) s.insert(*iter_s1);
  }
}

ElementId
OrderData_Tree::least_greatest(set<ElementId> ss, bool least)
{
  /*
    least m: for all x is (m, x), m, x from ss
    greatest M: for all x is (x, M), x, M from ss
  */
  set<ElementId>::iterator iter_s1, iter_s2;
  
  for (iter_s1 = ss.begin(); iter_s1 != ss.end(); iter_s1++) {
    for (iter_s2 = ss.begin(); iter_s2 != ss.end(); iter_s2++) {
      if (*iter_s1 == *iter_s2) continue;
      if (least) { if (!lessTransitive(*iter_s1, *iter_s2)) break; }
      else if (!lessTransitive(*iter_s2, *iter_s1)) break;
    }
    if (iter_s2 == ss.end()) return *iter_s1;
  }
  return NoElement;
}

void
OrderData_Tree::lower_upperCone(set<ElementId> &s, set<ElementId> ss, 
				bool lower)
{
  /*
    lower: all m such that for all x is (m, x), x from ss
    upper: all M such that for all x is (x, M), x from ss
  */
  set<Pair *, Pair_compare_first>::iterator iter_sp;
  set<ElementId>::iterator iter_se;
  
  for (iter_sp = pairs_first.begin(); (iter_sp != pairs_first.end()) && 
	 ((*iter_sp)->first == 0); iter_sp++) {
    for (iter_se = ss.begin(); iter_se != ss.end(); iter_se++) {
      if ((*iter_sp)->second == *iter_se) continue;
      if (lower) { 
	if (!lessTransitive((*iter_sp)->second, *iter_se)) break;
      }
      else if (!lessTransitive(*iter_se, (*iter_sp)->second)) break;
    }
    if (iter_se == ss.end()) s.insert((*iter_sp)->second);
  }
}

void
OrderData_Tree::getPaths_(list<set<ElementId> > &l, set<ElementId> s, 
			  ElementId x, ElementId y)
{
  Pair pr;
  set<Pair *, Pair_compare_first>::iterator iter_s;
  bool transitive = false;
  pair<set<ElementId>::iterator, bool> p;

  pr.first = x; pr.second = NoElement;
  iter_s = pairs_first.find(&pr);
  if (iter_s == pairs_first.end()) return;
  for (iter_s++; (iter_s != pairs_first.end()) && ((*iter_s)->first == x); 
       iter_s++) {
    if (!lessTransitive((*iter_s)->second, y, true)) continue;
    transitive = true;
    p = s.insert((*iter_s)->second);
    getPaths_(l, s, (*iter_s)->second, y);
    s.erase(p.first);
  }
  if (!transitive) l.push_back(s);
}

OrderData_Tree::~OrderData_Tree()
{
  clear();
}

OrderData *
OrderData_Tree::getCopy()
{
  OrderData_Tree *copy;
  Pair *new_pair;
  set<Pair *, Pair_compare_first>::iterator iter_spf;

  copy = new OrderData_Tree();
  if (!copy) return NULL;
  for (iter_spf = pairs_first.begin(); iter_spf != pairs_first.end();
       iter_spf++) {
    new_pair = new Pair;
    new_pair->first = (*iter_spf)->first;
    new_pair->second = (*iter_spf)->second;
    copy->pairs_first.insert(new_pair);
    copy->pairs_second.insert(new_pair);
  }
  return copy;
}

void
OrderData_Tree::clear()
{
  set<Pair *, Pair_compare_first>::iterator iter_s;
  
  for (iter_s = pairs_first.begin(); iter_s != pairs_first.end(); iter_s++)
    delete *iter_s;
  pairs_first.clear();
  pairs_second.clear();
}

bool
OrderData_Tree::newElement(ElementId x)
{
  if (isElement(x)) return false;
  insertPair(x, NoElement);
  insertPair(NoElement, x);
  return true;
}

bool
OrderData_Tree::removeElement(ElementId x)
{
  /*
    remove also all pairs containing x
  */
  Pair p;
  ElementId a, b;
  set<Pair *, Pair_compare_first>::iterator iter_spf1, iter_spf2;
  set<Pair *, Pair_compare_second>::iterator iter_sps1, iter_sps2;

  p.first = x; p.second = NoElement;
  iter_spf1 = pairs_first.find(&p);
  if (iter_spf1 == pairs_first.end()) return false;
  do {
    b = (*iter_spf1)->second;
    iter_spf2 = iter_spf1;
    iter_spf1++;
    iter_sps1 = pairs_second.find(*iter_spf2);
    delete *iter_spf2;
    pairs_first.erase(iter_spf2); pairs_second.erase(iter_sps1);
    if (b == NoElement) continue;
    p.first = NoElement; p.second = x;
    for (iter_sps1 = pairs_second.find(&p), iter_sps1++;
	 (iter_sps1 != pairs_second.end()) && ((*iter_sps1)->second == x);
	 iter_sps1++) {
      a = (*iter_sps1)->first;
      if (!(lessTransitive(b, a) || lessTransitive(a, b))) 
	insertPair(a, b);
    }
  } while ((iter_spf1 != pairs_first.end()) && ((*iter_spf1)->first == x));
  p.first = NoElement; p.second = x;
  iter_sps1 = pairs_second.find(&p);
  do {
    iter_sps2 = iter_sps1;
    iter_sps1++;
    iter_spf1 = pairs_first.find(*iter_sps2);
    delete *iter_sps2;
    pairs_second.erase(iter_sps2); pairs_first.erase(iter_spf1);
  } while ((iter_sps1 != pairs_second.end()) && ((*iter_sps1)->second == x));
  return true;
}

bool
OrderData_Tree::orderElements(ElementId x, ElementId y)
{
  Pair p;
  set<Pair *, Pair_compare_first>::iterator iter_spf;
  set<Pair *, Pair_compare_second>::iterator iter_sps;
  
  if ((!isElement(x)) || (!isElement(y))) return false;
  if (x == y) return false;
  if (lessTransitive(y, x) || lessTransitive(x, y)) return false;
  insertPair(x, y);
  p.first = y; p.second = NoElement;
  for (iter_spf = pairs_first.find(&p), iter_spf++;
       (iter_spf != pairs_first.end()) && ((*iter_spf)->first == y);
       iter_spf++)
    erasePair(x, (*iter_spf)->second);
  p.first = NoElement; p.second = x;
  for (iter_sps = pairs_second.find(&p), iter_sps++;
       (iter_sps != pairs_second.end()) && ((*iter_sps)->second == x);
       iter_sps++)
    erasePair((*iter_sps)->first, y);
  return true;
}

bool
OrderData_Tree::removeOrder(ElementId x, ElementId y)
{
  Pair p;
  ElementId a, b;
  set<Pair *, Pair_compare_first>::iterator iter_spf;
  set<Pair *, Pair_compare_second>::iterator iter_sps;
  
  if (x == y) return false;
  if (!erasePair(x, y)) return false;
  p.first = y; p.second = NoElement;
  for (iter_spf = pairs_first.find(&p), iter_spf++;
       (iter_spf != pairs_first.end()) && ((*iter_spf)->first == y);
       iter_spf++) {
    b = (*iter_spf)->second;
    if (!(lessTransitive(b, x) || lessTransitive(x, b)))
      insertPair(x, b);
  }
  p.first = NoElement; p.second = x;
  for (iter_sps = pairs_second.find(&p), iter_sps++;
       (iter_sps != pairs_second.end()) && ((*iter_sps)->second == x);
       iter_sps++) {
    a = (*iter_sps)->first;
    if (!(lessTransitive(y, a) || lessTransitive(a, y)))
      insertPair(a, y);
  }
  return true;
}

bool
OrderData_Tree::less(ElementId x, ElementId y, bool transitivity)
{
  if ((!isElement(x)) || (!isElement(y))) return false;
  if (x == y) return true;
  return lessTransitive(x, y, transitivity);
}

void
OrderData_Tree::getElements(set<ElementId> &s)
{
  set<Pair *, Pair_compare_first>::iterator iter_s;
  
  for (iter_s = pairs_first.begin(); (iter_s != pairs_first.end()) && 
	 ((*iter_s)->first == NoElement); iter_s++)
    s.insert((*iter_s)->second);
}

void
OrderData_Tree::upperElements(set<ElementId> &s, ElementId x)
{
  Pair p;
  set<Pair *, Pair_compare_first>::iterator iter_s;
  
  p.first = x; p.second = NoElement;
  iter_s = pairs_first.find(&p);
  if (iter_s == pairs_first.end()) return;
  for (iter_s++; (iter_s != pairs_first.end()) && ((*iter_s)->first == x); 
       iter_s++)
    s.insert((*iter_s)->second);
}

void
OrderData_Tree::lowerElements(set<ElementId> &s, ElementId x)
{
  Pair p;
  set<Pair *, Pair_compare_second>::iterator iter_s;
  
  p.first = NoElement; p.second = x;
  iter_s = pairs_second.find(&p);
  if (iter_s == pairs_second.end()) return;
  for (iter_s++; (iter_s != pairs_second.end()) && ((*iter_s)->second == x); 
       iter_s++)
    s.insert((*iter_s)->first);
}

void
OrderData_Tree::minimalElements(set<ElementId> &s, set<ElementId> ss)
{
  minimal_maximal(s, ss, true);
}

void
OrderData_Tree::maximalElements(set<ElementId> &s, set<ElementId> ss)
{
  minimal_maximal(s, ss, false);
}

ElementId
OrderData_Tree::leastElement(set<ElementId> ss)
{
  return least_greatest(ss, true);
}

ElementId
OrderData_Tree::greatestElement(set<ElementId> ss)
{
  return least_greatest(ss, false);
}

void
OrderData_Tree::lowerCone(set<ElementId> &s, set<ElementId> ss)
{
  lower_upperCone(s, ss, true);
}

void
OrderData_Tree::upperCone(set<ElementId> &s, set<ElementId> ss)
{
  lower_upperCone(s, ss, false);
}

ElementId
OrderData_Tree::infimum(set<ElementId> ss)
{
  set<ElementId> s1;
  
  lower_upperCone(s1, ss, true);
  return least_greatest(s1, false);
}

ElementId
OrderData_Tree::supremum(set<ElementId> ss)
{
  set<ElementId> s1;

  lower_upperCone(s1, ss, false);
  return least_greatest(s1, true);
}

void
OrderData_Tree::getPaths(list<set<ElementId> > &l, ElementId x,
			 ElementId y)
{
  set<ElementId> s;
  
  if (!lessTransitive(x, y, true)) return;
  s.insert(x); s.insert(y);
  getPaths_(l, s, x, y);
}
