/* LatVis
 * Copyright (C) 2002 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "orderedset.h"
#include "orderdata_tree.h"
#include "orderdata_list.h"
#include "orderdata_array.h"
#include "diagram_graphic.h"
#include <stdio.h>

OrderedSet::UndoRedoType *
OrderedSet::getState()
{
  UndoRedoType *state = NULL;

  state = new UndoRedoType;
  if (!state) return NULL;
  state->elements = elements;
  state->data = data->getCopy();
  state->diagram = diagram->getCopy(state->data);
  state->changed = changed; state->saved = saved;
  state->filename = filename;
  return state;
}

void
OrderedSet::setState(UndoRedoType *state)
{
  if (!state) return;
  elements = state->elements;
  delete data; data = state->data->getCopy();
  delete diagram; diagram = state->diagram->getCopy(data);
  changed = state->changed; saved = state->saved;
  filename = state->filename;
}

Element *
OrderedSet::getElementFromElementId(ElementId x)
{
  Element *retval = NULL, e(x);
  set<Element>::iterator iter_s;
  
  iter_s = elements.find(e);
  if ((x != NoElement) && (iter_s != elements.end())) {
    retval = new Element(NoElement);
    *retval = *iter_s;
  }
  return retval;
}

void 
OrderedSet::getElementsFromElementIds(set<Element> &s, set<ElementId> ss)
{
  set<ElementId>::iterator iter_s; 
  Element e(NoElement);
  set<Element>::iterator iter_s1;
  
  for (iter_s = ss.begin(); iter_s != ss.end(); iter_s++) {
    e.id = *iter_s;
    iter_s1 = elements.find(e);
    if (iter_s1 != elements.end()) s.insert(*iter_s1);
  }
}

void
OrderedSet::getElementIdsFromElements(set<ElementId> &s, set<Element> ss)
{
  set<Element>::iterator iter_s; 
  
  for (iter_s = ss.begin(); iter_s != ss.end(); iter_s++)
    s.insert((*iter_s).id);
}

OrderedSet::OrderedSet(data_t data_type, OrderedSetId set_id)
  : undo_list(true)
{
  data = NULL;
  datatype = data_type;
  switch (data_type) {
  case DATA_TREE: data = new OrderData_Tree(); break;
  case DATA_LIST: data = new OrderData_List(); break;
  case DATA_ARRAY: data = new OrderData_Array();
  }
  if (!data) return;
  diagram = NULL;
  diagram = new Diagram_Graphic(this, data, datatype);
  if (!diagram) { delete data; return; }
  changed = false; saved = true;
  filename = "";
  id = set_id;
}

OrderedSet::~OrderedSet()
{
  delete data;
  delete diagram;
}

void
OrderedSet::clear()
{
  elements.clear();
  data->clear();
  diagram->clearDiagram();
}

void
OrderedSet::save_UndoRedo()
{
  UndoRedoType *state = NULL;
  
  state = getState();
  if (!state) return;
  undo_list.save(state);
}

void
OrderedSet::removeLast_UndoRedo()
{
  undo_list.removeLast();
}

void
OrderedSet::undo()
{
  UndoRedoType *state = NULL, *new_state;
  
  state = getState();
  if (!state) return;
  new_state = undo_list.undo(state);
  if (state != new_state) setState(new_state);
}

void
OrderedSet::redo()
{
  UndoRedoType *state = NULL, *new_state;
  
  state = getState();
  if (!state) return;
  new_state = undo_list.redo(state);
  if (state != new_state) { setState(new_state); /*delete new_state;*/ }
}

Element *
OrderedSet::newElement(ElementName x, Vertex *v)
{
  set<Element>::iterator iter_s;
  Element *new_element;
  
  for (iter_s = elements.begin(); iter_s != elements.end(); iter_s++)
    if ((*iter_s).name == x) return NULL;
  new_element = new Element(1);
  for (iter_s = elements.begin(); iter_s != elements.end();
       iter_s++, new_element->id++)
    if ((*iter_s).id != new_element->id) break;
  new_element->name = x;
  if (v) v->element = new_element->id;
  if ((!data->newElement(new_element->id)) ||
      ((v) && (!(v = diagram->addVertex(*v))))) {
    data->removeElement(new_element->id);
    if (v) delete v;
    delete new_element;
    return NULL;
  }
  if (v) delete v;
  elements.insert(*new_element);
  setChanged();
  return new_element;
}

bool
OrderedSet::rename(Element x, ElementName name)
{
  set<Element>::iterator iter_s;
  Element elem(NoElement);

  for (iter_s = elements.begin(); iter_s != elements.end(); iter_s++)
    if ((*iter_s).name == name) return false;
  iter_s = elements.find(x);
  if (iter_s == elements.end()) return false;
  elem = *iter_s; elem.name = name;
  elements.erase(iter_s);
  elements.insert(elem);
  setChanged();
  return true;
}

bool
OrderedSet::removeElement(Element x)
{
  set<Element>::iterator iter_s;
  Vertex v(x.id);
  
  iter_s = elements.find(x);
  if (iter_s == elements.end()) return false;
  if (!data->removeElement(x.id)) return false;
  diagram->deleteVertex(v);
  elements.erase(iter_s);
  setChanged();
  return true;
}

bool
OrderedSet::orderElements(Element x, Element y)
{
  Vertex *vx, *vy;
  ElementId id1, id2;
  
  id1 = x.id; id2 = y.id;
  vx = diagram->getVertex(x);
  vy = diagram->getVertex(y);
  if (vx && vy && (vx->y > vy->y)) { 
    delete vx; delete vy;
    id1 = y.id; id2 = x.id;
  }
  if (!data->orderElements(id1, id2)) return false;
  setChanged();
  return true;
}

bool
OrderedSet::removeOrder(Element x, Element y)
{
  Vertex *vx, *vy;
  ElementId id1, id2;

  id1 = x.id; id2 = y.id;
  vx = diagram->getVertex(x);
  vy = diagram->getVertex(y);
  if (vx && vy && (vx->y > vy->y)) { 
    delete vx; delete vy;
    id1 = y.id; id2 = x.id;
  }
  if (!data->removeOrder(id1, id2)) return false;
  setChanged();
  return true;
}

bool
OrderedSet::less(Element x, Element y, bool transitivity)
{
  return data->less(x.id, y.id, transitivity);
}

void
OrderedSet::getElements(set<Element> &s)
{
  //set<ElementId> s_;
  
  //data->getElements(s_);
  //getElementsFromElementIds(s, s_);
  s = elements;
}

void
OrderedSet::upperElements(set<Element> &s, Element x)
{
  set<ElementId> s_;
  
  data->upperElements(s_, x.id);
  getElementsFromElementIds(s, s_);
}

void
OrderedSet::lowerElements(set<Element> &s, Element x)
{
  set<ElementId> s_;
  
  data->lowerElements(s_, x.id);
  getElementsFromElementIds(s, s_);
}

void
OrderedSet::minimalElements(set<Element> &s, set<Element> ss)
{
  set<ElementId> s_, ss_;

  getElementIdsFromElements(ss_, ss);
  data->minimalElements(s_, ss_);
  getElementsFromElementIds(s, s_);
}

void
OrderedSet::maximalElements(set<Element> &s, set<Element> ss)
{
  set<ElementId> s_, ss_;

  getElementIdsFromElements(ss_, ss);
  data->maximalElements(s_, ss_);
  getElementsFromElementIds(s, s_);
}

Element *
OrderedSet::leastElement(set<Element> ss)
{
  set<ElementId> ss_;
  
  getElementIdsFromElements(ss_, ss);
  return getElementFromElementId(data->leastElement(ss_));
}

Element *
OrderedSet::greatestElement(set<Element> ss)
{
  set<ElementId> ss_;
  
  getElementIdsFromElements(ss_, ss);
  return getElementFromElementId(data->greatestElement(ss_));
}

void
OrderedSet::lowerCone(set<Element> &s, set<Element> ss)
{
  set<ElementId> s_, ss_;

  getElementIdsFromElements(ss_, ss);
  data->lowerCone(s_, ss_);
  getElementsFromElementIds(s, s_);
}

void
OrderedSet::upperCone(set<Element> &s, set<Element> ss)
{
  set<ElementId> s_, ss_;

  getElementIdsFromElements(ss_, ss);
  data->upperCone(s_, ss_);
  getElementsFromElementIds(s, s_);
}

Element *
OrderedSet::infimum(set<Element> ss)
{
  set<ElementId> ss_;

  getElementIdsFromElements(ss_, ss);
  return getElementFromElementId(data->infimum(ss_));
}

Element *
OrderedSet::supremum(set<Element> ss)
{
  set<ElementId> ss_;

  getElementIdsFromElements(ss_, ss);
  return getElementFromElementId(data->supremum(ss_));
}

void
OrderedSet::getPaths(list<set<Element> > &l, Element x, Element y)
{
  list<set<ElementId> > l_;
  list<set<ElementId> >::iterator iter_l;
  set<Element> s;
  Vertex *vx, *vy;
  ElementId id1, id2;

  id1 = x.id; id2 = y.id;
  vx = diagram->getVertex(x);
  vy = diagram->getVertex(y);
  if (vx && vy && (vx->y > vy->y)) { 
    delete vx; delete vy;
    id1 = y.id; id2 = x.id;
  }
  data->getPaths(l_, id1, id2);
  for (iter_l = l_.begin(); iter_l != l_.end(); iter_l++) {
    s.clear();
    getElementsFromElementIds(s, *iter_l);
    l.push_back(s);
  }
}

Element *
OrderedSet::getElement(ElementName x)
{
  set<Element>::iterator iter_s;
  
  for (iter_s = elements.begin(); iter_s != elements.end(); iter_s++)
    if ((*iter_s).name == x) return getElementFromElementId((*iter_s).id);  
  return NULL;
}
